/* $Modified: Wed Jul  7 13:55:05 2010 by uwer $ */
// vim: set ft=cpp:
// vim: set foldmethod=marker:
#ifndef COEFFICIENTS_H_
#define COEFFICIENTS_H_
#include <iostream>
#include <complex>
#include <cmath>

// turn off NDEBUG to bypass assert statements
#define NDEBUG
#include <assert.h>


#include "mytypes.h"
//{{{

struct Topology { 
  int h0;
  int h1;
  int h2;
  int h3;
  int h4;
  std::complex<DOUBLE> l0[4];
  std::complex<DOUBLE> l1[4];
  std::complex<DOUBLE> l2[4];
  std::complex<DOUBLE> l3[4];
  std::complex<DOUBLE> l4[4];
  std::complex<DOUBLE> scalarMSq;

};

class Coefficients{

 public:
  ~Coefficients(){
    delete [] C4;
    delete [] C3;
    delete [] C2;
    delete [] R5;
    delete [] R4;
    delete [] R3;
    delete [] R2;
    delete [] momC4;
    delete [] momC3;
    delete [] momC2;
  };
  
 Coefficients(int NGLUON_) : n(NGLUON_) {
    
    int size;
    
    size = n*(n-1)*(n-2)*(n-3)/24; // bino(n,4)
    momC4  = new std::complex<DOUBLE>[2*size][4];
    C4  = new std::complex<DOUBLE> [2*size];
    R4  = new std::complex<DOUBLE> [6*size];
    
    size = n*(n-1)*(n-2)/6; // bino(n,3);
    momC3  = new std::complex<DOUBLE> [3*size][4];
    C3  = new std::complex<DOUBLE> [7*size];
    R3  = new std::complex<DOUBLE> [10*size];
    
    size = (n-1)*(n-2)/2; // bino(n-1,2)
    C2  = new std::complex<DOUBLE> [size];
    R2  = new std::complex<DOUBLE> [size];
    momC2  = new std::complex<DOUBLE>[4*size][4];

    size = n*(n-1)*(n-2)*(n-3)*(n-4)/120; // bino(n,5)
    R5  = new std::complex<DOUBLE> [size];
    M5  = new DOUBLE [size];
  }
  
  int penkey(const int i1, const int i2, const int i3, const int i4,const int i5){
    assert(i5>i4 && i4>i3 && i3>i2 && i2>i1);
    return(i1+i2*(i2-1)/2+i3*(i3-1)*(i3-2)/6+i4*(i4-1)*(i4-2)*(i4-3)/24
	   +i5*(i5-1)*(i5-2)*(i5-3)*(i5-4)/120);
  }
  
  int boxkey(const int  i1,const int  i2, const int  i3, const int  i4){
    assert( i4>i3 && i3>i2 && i2>i1 );
    return( i1 + i2*(i2-1)/2 + i3*(i3-1)*(i3-2)/6 + i4*(i4-1)*(i4-2)*(i4-3)/24 );
  }
  
  int trikey(const int  i1,const int  i2,const int  i3){
    assert(i3>i2 && i2>i1);
    return( i1+i2*(i2-1)/2+i3*(i3-1)*(i3-2)/6);
  }
  
  int bubkey(const int  i1,const int  i2){ 
    assert(i2>i1);
    return(i1+(i2-1)*(i2-2)/2);
  }
  
  void setbox(const int  key,int i5, const std::complex<DOUBLE> & coeff){ 
    C4[2*key+i5]=coeff;
  }
  
  std::complex<DOUBLE> box(const int  key, int i5){
    return C4[2*key+i5];
  }
  
  void setboxR(const int  key, const int  i5, const std::complex<DOUBLE> & coeff){
    R4[6*key+i5]=coeff;
  }
  
  std::complex<DOUBLE> boxR(const int  key, const int  i5){
    return R4[6*key+i5];
  }
  
  void setboxmom(const int  key, const int  i5, const std::complex<DOUBLE> mom[4]){
    int i = 2*key+i5;
    for (int mu = 0; mu < 4; mu++){
      momC4[i][mu] = mom[mu];
    }
  }
  
  std::complex<DOUBLE>* boxmom(const int  key, const int  i5){
    return momC4[2*key+i5];
  }
  
  void settri(const int  key, const int  i4, const std::complex<DOUBLE> & coeff){
    C3[i4+7*key]=coeff;
  }
  
  void settriR(const int  key, const int  i4, const std::complex<DOUBLE> & coeff){
    R3[i4+10*key]=coeff;
  }
  
  void settrimom(const int key,const int i4, const std::complex<DOUBLE> mom[4]){
    int i = i4+3*key;
    for (int mu = 0; mu < 4; mu++){
      momC3[i][mu] = mom[mu];
    }
  }
  
  std::complex<DOUBLE> tri(const int  key, const int  i4){
    return C3[i4+7*key];
  }
  
  std::complex<DOUBLE> triR(const int  key, const int  i4){
    return R3[i4+10*key];
  }
  
  std::complex<DOUBLE>* trimom(const int  key,const int  i4){
    return momC3[i4+3*key];
  }
  
  
  void setbub(const int  key, const int  i3, const std::complex<DOUBLE> & coeff){
    C2[i3+key]=coeff;
  }
  
  void setbubR(const int  key, const int  i3, const std::complex<DOUBLE> & coeff){
    R2[i3+key]=coeff;
  }
  
  void setbubmom(const int  key, const int  i3, const std::complex<DOUBLE> mom[4]){
    for (int mu = 0; mu < 4; mu++){
      momC2[i3+4*key][mu] = mom[mu];
    }
  }
  
  std::complex<DOUBLE> bub(const int  key, const int  i3){
    return C2[i3+key];
  }
  
  std::complex<DOUBLE> bubR(const int  key, const int  i3){
    return R2[i3+key];
  }
  
  std::complex<DOUBLE>* bubmom(const int  key, const int  i3){
    return momC2[i3+4*key];
  }
  

  void setpentagonR(const int  key, const std::complex<DOUBLE>& coeff){
    R5[key]=coeff;
  }
  
  std::complex<DOUBLE> pentagonR(const int key){
    return R5[key];
  }
  
  void setpentagonM(const int  key, const DOUBLE scalarMass){
    M5[key]=scalarMass;
  }
  
  DOUBLE pentagonM(const int key){
    return M5[key];
  }

 private:
  int n;
  std::complex<DOUBLE> *C4, *C3, *C2, *R5, *R4, *R3, *R2;
  DOUBLE *M5;
  std::complex<DOUBLE> (*momC4)[4], (*momC3)[4], (*momC2)[4];
  
};
//}}}
#endif // COEFFICIENTS_H_
