// $Modified: Wed Nov  3 14:00:45 2010 by uwer $
#include <iostream>
#include "Current.h"
#define NDEBUG
#include <assert.h>

#include <cstdlib>

using namespace std;


#define DOTP(X,Y) ( X[0]*Y[0] - ( X[1]*Y[1] + X[2]*Y[2] + X[3]*Y[3] ) )


void Current::setRefMom(const DOUBLE* ref_){
  for (int mu=0; mu<4; mu++){
    refMom[mu] = ref_[mu];
  }
}


template<typename T>
void printMomentum(const T *mom){

  cout << "mom = ";
  for (int mu=0; mu<4; mu++){
    cout << mom[mu] << ",  ";
  }
  cout << endl;
}

template void printMomentum(const DOUBLE *mom);
template void printMomentum(const complex<DOUBLE> *mom);

void Current::cyclicMomSum(){
  for (int i=1; i < ngluon; i++){
    for (int j=0; j < ngluon; j++){
      for (int mu=0; mu<4; mu++){
	KKK[i][j][mu] 
	  = KKK[i-1][j][mu] + KKK[0][((j+i-1) % ngluon+1) % ngluon][mu];
      }
      S_ij[i][j] = DOTP(KKK[i][j],KKK[i][j]);
    }
  }
}

void Current::cyclicSum( int Matrix[NMAX][NMAX] ){
  for (int i=1; i < ngluon; i++){
    for (int j=0; j < ngluon; j++){
      Matrix[i][j] 
	= Matrix[i-1][j] + Matrix[0][((j+i-1) % ngluon+1) % ngluon];
    }
  }
}


void Current::momentumUpdate(){
  for(int i=0; i<ngluon; i++){
    S_ij[0][i] = zero;
    for(int mu=0; mu<4; mu++){
      KKK[0][i][mu] = k[i][mu];
    }
  }
  Current::cyclicMomSum();
}


void Current::helicityUpdate(){
  for(int i=0; i<ngluon; i++){
    Helsum[0][i] = hel[i];
  }
  Current::cyclicSum(Helsum);
}

string Current::helicity2string(const int n, int hel[]){

  string helString;

  for (int i=0; i<n; i++){
    if (hel[i]==-1)
      helString += "-";
    else
      helString += "+";
  }
  return helString;
}

string Current::int2string(int number){

   stringstream ss;
   ss << number;
   return ss.str();
}

void Current::sethel(unsigned int n, int htype, int *helicity){
  int one(1);
  switch (htype){
  case 0:
    for(unsigned int h=0;h<n;h++)
      helicity[h] = one;
    break;
  case 1:
    helicity[0]=-one;
    for(unsigned int h=1;h<n;h++)
      helicity[h] = one;
    break;
  case 2:
    helicity[0]=-one;
    helicity[1]=-one;
    for(unsigned int h=2;h<n;h++)
      helicity[h] = one;
    break;
  case 3:
    for(unsigned int h=0;h<n;h++){
      if( h & 1 )
	helicity[h] = one;
      else
	helicity[h] = -one;
    }
    break;
  case 4:
    for(unsigned int h=0;h<n;h++){
      if( h & 1 )
	helicity[h] = -one;
      else
	helicity[h] = one;
    }
    break;
  default:
    cout << "Helicty configuration " << helicity2string(n,helicity) 
	 << " not supported \n";
  }
} 

void Current::sethel(unsigned int n, string hstr, int *helicity){ 

  int one(1);

  if ( hstr.length() == 1 ){
    istringstream in(hstr);
    int ihel;
    in >> ihel;
    sethel(n,ihel,helicity);
  }    

  if ( hstr.length() == n ){
    for(unsigned int h=0;h<n;h++){
      if(hstr.substr(h,1).compare("+")==0){ 
	helicity[h]=one;
      }
      else 
	if(hstr.substr(h,1).compare("-")==0){
	  helicity[h]=-one;
	}
	else { 
	  cout << "Unrecognised helicity " << hstr.substr(h,1) << endl;
	  cout << "use +/- to denote positive and negative helicities" << endl;
	  exit (1);
	}
    }
  } 
  
  if ( hstr.length()!=1 && hstr.length()!=n ){
    cout << "Length of helicity string does not match number of gluons." << endl;
    exit (1);
  }
  
}

void Current::rescaleMomenta(const DOUBLE x){

  for (int j=0; j<ngluon; j++){
    for(int mu=0; mu<4; mu++){
      KKK[0][j][mu] *= x;
    }
  }
  const DOUBLE xx = x*x;
  for (int i=1; i<ngluon; i++){
    for (int j=0; j<ngluon; j++){
      S_ij[i][j] *= xx;
      for(int mu=0; mu<4; mu++){
	KKK[i][j][mu] *= x;
      }
    }
  }
}


void Current::printMomenta(const int N){
  cout << " Momenta: number of Gluons = " << ngluon << endl;
  cout << "          index line       = " << N << endl;
  for(int a=0; a<ngluon; a++){
    for(int b=0; b<4; b++){
      cout <<" "<< KKK[N][a][b]<<",\t";
    }
    cout << endl;
  }
}


void Current::printCurrents(const int N){
  cout << " Currents: number of Gluons = " << ngluon << endl;
  cout << "           index line       = " << N << endl;
  for(int a=0; a<ngluon; a++){
    for(int b=0; b<4; b++){
      cout <<" "<< JJJ[N][a][b]<<",\t";
    }
    cout << endl;
  }
}

template<typename T>
void Current::get1PtCurrents(const int helicity, complex<DOUBLE> JJ[4],
			     const T KK[4]){

  const DOUBLE dhel = (helicity == 1) ? one : -one;
  // matched such that it agrees with the references from Giele-Zanderighi
  complex<DOUBLE> t1,t10,t13,t15,t19,t2,t3,t4,t5,t6,t7,t8,t20;

//  t1 = sqrt(two);
//  t2 = KK[1]-I*KK[3] * dhel; 
//  t3 = t1*t2;
//  t4 = refMom[0]-refMom[2];
//  t5 = t4*t2;
//  t6 = KK[0]+KK[2];
//  t7 = refMom[1]-I*refMom[3] *dhel;
//  t8 = t6*t7;
//  t10 = KK[0]-KK[2];
//  t13 = one/(t5-t10*t7);
//  t15 = one/t6;
//  t19 = t6*t4;
//  t20 = t2*t7;
//  JJ[0] = t3*(t5+t8)*t13*t15/two;
//  JJ[1] = t3*(t19+t20)*t13*t15/two;
//  JJ[2] = t3*(-t5+t8)*t13*t15/two;
//  JJ[3] = -I*t1*t2*(t19-t20)*t13*t15/two *dhel;

  t1 = sqrt(two);
  t2 = KK[3]-I*KK[1] * dhel; 
  t3 = t1*t2;
  t4 = refMom[0]-refMom[2];
  t5 = t4*t2;
  t6 = KK[0]+KK[2];
  t7 = refMom[3]-I*refMom[1] *dhel;
  t8 = t6*t7;
  t10 = KK[0]-KK[2];
  t13 = one/(t5-t10*t7);
  t15 = one/t6;
  t19 = t6*t4;
  t20 = t2*t7;
  JJ[0] = t3*(t5+t8)*t13*t15/two;
  JJ[3] = t3*(t19+t20)*t13*t15/two;
  JJ[2] = t3*(-t5+t8)*t13*t15/two;
  JJ[1] = -I*t1*t2*(t19-t20)*t13*t15/two *dhel;
 
}

template void Current::get1PtCurrents(const int helicity,
				      complex<DOUBLE> JJ[4],
				      const DOUBLE KK[4]);
template void Current::get1PtCurrents(const int helicity,
				      complex<DOUBLE> JJ[4],
				      const complex<DOUBLE> KK[4]);


inline void Current::vertex3GGG(const int len, const int k,
			       const int  m, const int l){

  DOUBLE *K_k = posX(k, m, KKK);
  DOUBLE *K_l = posX((m+1)%ngluon, l, KKK);
  complex<DOUBLE> *J_k = posX(k, m, JJJ);
  complex<DOUBLE> *J_l = posX((m+1)%ngluon, l, JJJ);

  const DOUBLE two(2.);
  const complex<DOUBLE> KlxJk(two*DOTP(K_l, J_k));
  const complex<DOUBLE> KkxJl(two*DOTP(K_k, J_l));
  const complex<DOUBLE> JkxJl(DOTP(J_k, J_l));
  for (unsigned int i=0; i<4; i++){
    JJJ[len][k][i] +=  KlxJk * J_l[i] - KkxJl * J_k[i] 
      + JkxJl * ( K_k[i] - K_l[i] );
  }

}


template<typename T>
  inline void Current::vertex3GGG_c(const int dest, const int k, const int x,
				    const  int l, T KK[][4],
				    complex<DOUBLE> JJ[][4]){

  T *K_k = KK[k];
  DOUBLE *K_l = posX(x, l, KKK);
  complex<DOUBLE> *J_k = JJ[k];
  complex<DOUBLE> *J_l = posX(x, l, JJJ);

  const DOUBLE two(2.);
  const complex<DOUBLE> KlxJk(two*DOTP(K_l, J_k));
  const complex<DOUBLE> KkxJl(two*DOTP(K_k, J_l));
  const complex<DOUBLE> JkxJl(DOTP(J_k, J_l));
  for (unsigned int i=0; i<4; i++){
    JJ[dest][i] +=  KlxJk * J_l[i] - KkxJl * J_k[i] 
      + JkxJl * ( K_k[i] - K_l[i] );
  }
}


template<typename T>
  inline void Current::vertex3GSS_c(const int dest, const int k, const int x,
				    const  int l, T KK[][4],
				    complex<DOUBLE> JJ[]){

  T *K_k = KK[k];
  // DOUBLE *K_l = posX(x, l, KKK);
  complex<DOUBLE> J_k = JJ[k];
  complex<DOUBLE> *J_l = posX(x, l, JJJ);

  const complex<DOUBLE> KkxJl(DOUBLE(2.)*DOTP(K_k, J_l));
  
  JJ[dest] +=  - KkxJl * J_k;
}


inline void Current::vertex4GGGG(const int len, const int k, const int  m, 
				const int j, const int l){

  complex<DOUBLE> *J_k = posX(k,m,JJJ);
  complex<DOUBLE> *J_s = posX((m+1)%ngluon,j,JJJ);
  complex<DOUBLE> *J_l = posX((j+1)%ngluon,l,JJJ);
  
  const complex<DOUBLE> JkxJl(DOTP(J_k, J_l));
  const complex<DOUBLE> JkxJs(DOTP(J_k, J_s));
  const complex<DOUBLE> JlxJs(DOTP(J_l, J_s));
  for (unsigned int i=0; i<4; i++){
    JJJ[len][k][i] += DOUBLE(2.)*JkxJl*J_s[i] - JkxJs*J_l[i] - JlxJs*J_k[i];
  }
 
}


inline void Current::vertex4GGGG_c(const int len, const int k, const int x, 
				   const int y, const int l,
				   complex<DOUBLE> JJ[][4]){
  
  complex<DOUBLE> *J_k = JJ[k];
  complex<DOUBLE> *J_s = posX(x,y-1,JJJ);
  complex<DOUBLE> *J_l = posX(y,l,JJJ);
  
  const complex<DOUBLE> JkxJl(DOTP(J_k, J_l));
  const complex<DOUBLE> JkxJs(DOTP(J_k, J_s));
  const complex<DOUBLE> JlxJs(DOTP(J_l, J_s));
  for (unsigned int i=0; i<4; i++){
    JJ[len][i] += DOUBLE(2.)*JkxJl*J_s[i] - JkxJs*J_l[i] - JlxJs*J_k[i];
  }
  
}


inline void Current::vertex4GGSS_c(const int len, const int k, const int x, 
				   const int y, const int l,
				   complex<DOUBLE> JJ[]){
  
  complex<DOUBLE> J_k = JJ[k];
  complex<DOUBLE> *J_s = posX(x,y-1,JJJ);
  complex<DOUBLE> *J_l = posX(y,l,JJJ);
  
  const complex<DOUBLE> JlxJs(DOTP(J_l, J_s));

  JJ[len] +=  - JlxJs*J_k;

}


void Current::fillPartialCurrents(){

  for (int i=0; i<ngluon; i++){
      get1PtCurrents(hel[i], JJJ[0][i], KKK[0][i]);
  }

  // toDo: we need it only until len<n-2;

  // len counts the number of external momenta involved
  for (int len=1; len < ngluon-1; len++){ 

    // k is the first momentum index, l is the last momentum index
    for (int k=0; k < ngluon; k++){
      const int l_virtual = k+len;
      const int l = (l_virtual) % ngluon;

      for(unsigned int j=0; j<4; j++){
	JJJ[len][k][j] = 0.;
      }

      // cyclic momentum sum
      for (unsigned int mu=0; mu<4; mu++){
	KKK[len][k][mu] 
	  = KKK[len-1][k][mu] + KKK[0][((k+len-1)%ngluon+1)%ngluon][mu];
      }

      Helsum[len][k] = Helsum[len-1][k] + Helsum[0][((k+len-1)%ngluon+1)%ngluon];

      // three gluon vertex
      for (int m=k; m < l_virtual; m++){
	vertex3GGG(len, k, m % ngluon, l);
      }

      // four gluon vertex
      for (int m=k; m < l_virtual-1; m++){
	for (int j=m+1; j < l_virtual; j++){
	  vertex4GGGG(len, k, m % ngluon, j % ngluon, l);
	}
      }

      // include the propagators
      if ( len < ngluon-2) {
	complex<DOUBLE> den = DOUBLE(1.)/S_ij[len][k];
	for (unsigned int mu=0; mu<4; mu++){
	  JJJ[len][k][mu] *= den;
	}
      } 
      
    }
  }
}


complex<DOUBLE> Current::partialGluonAmplitude(const int i, const int j,
					       const int hel_i,const int hel_j,
					       const complex<DOUBLE> l_i[4],
					       const complex<DOUBLE> l_j[4]){
  // local number of external gluons plus loop gluons 
  const int N = (j-i+ngluon) % ngluon + 3;

  assert(N>0);
  
  // prepare additional arrays
  complex<DOUBLE> KK[N][4];
  complex<DOUBLE> JJ[N][4];


  for(unsigned int mu=0; mu<4; mu++){
    KK[0][mu] = -l_i[mu]; 
  }

  get1PtCurrents(hel_i,JJ[0],KK[0]);
  get1PtCurrents(hel_j,JJ[N-1],l_j);


  for (int len=1; len<N-1; len++){//len=numbers of gluons in the current
    
    const int ipos = (i+len-1) % ngluon;

    for (unsigned int mu=0; mu<4; mu++){
      KK[len][mu] = KK[len-1][mu] + KKK[0][ipos][mu];
      JJ[len][mu] = 0.;
    }
    
    // three gluon vertex
    for ( int m=0; m<len; m++){
      vertex3GGG_c(len, m, (i+m)%ngluon, ipos, KK, JJ);
    }

    // four gluon vertex
    for (int m=0; m<len-1; m++){
      for (int mm=m+1; mm<len; mm++){
	vertex4GGGG_c(len, m, (i+m)%ngluon, (i+mm)%ngluon, ipos, JJ);
      }
    }

    // include the propagators
    if ( len < N-2) {
      complex<DOUBLE> den = DOUBLE(1.)
	/ ( DOUBLE(2.)*DOTP(KKK[len-1][i],KK[0]) + S_ij[len-1][i]);
      // corresponds to: den = DOUBLE(1.)/DOTP(KK[len],KK[len]);
      for (unsigned int mu=0; mu<4; mu++){
	JJ[len][mu] *= den;
      }
    }
  }
  
  return( DOTP(JJ[N-2],JJ[N-1]) * DOUBLE(2.) * I / pow(sqrt(DOUBLE(2.)),N) );
}


complex<DOUBLE> Current::partialGluonScalarAmplitude(const int i, const int j,
						     const complex<DOUBLE>  l_i[4],
						     const complex<DOUBLE>  l_j[4]){

  // local number of external gluons plus loop scalars 
  const int N = (j-i+ngluon) % ngluon + 3;

  assert(N>0);
  
  // prepare additional arrays
  complex<DOUBLE> KK[N][4];
  complex<DOUBLE> JJ[N];

  for(unsigned int mu=0; mu<4; mu++){
    KK[0][mu] = -l_i[mu]; 
  }
  JJ[0] = one;
  // JJ[N-1] = one; not neccessary: multiplication with 1 only in the last step

  for (int len=1; len<N-1; len++){//len= # of particles in the current

    for (unsigned int mu=0; mu<4; mu++){
      KK[len][mu] = KK[len-1][mu] + KKK[0][(i+len-1)%ngluon][mu];
    }
    JJ[len] = 0.;
 
    // three vertex
     for ( int m=0; m<len; m++){
      vertex3GSS_c(len, m, (i+m)%ngluon, (i+len-1)%ngluon, KK, JJ);
    }
 
    // four vertex
    for (int m=0; m<len-1; m++){
      for (int mm=m+1; mm<len; mm++){
	vertex4GGSS_c(len, m, (i+m)%ngluon, (i+mm)%ngluon, (i+len-1)%ngluon, JJ);
      }
    }
    

    // include the propagators
    if ( len < N-2) {
      JJ[len] /= ( DOUBLE(2.) * DOTP(KKK[len-1][i],KK[0]) + S_ij[len-1][i] );
      // corresponds to: JJ[len] /= ( DOTP(KK[len],KK[len]) - scalarMassSq );
    }
  }

  return( JJ[N-2] /* *JJ[N-1]*/* DOUBLE(2.) * I/pow(sqrt(DOUBLE(2.)),N));
  // JJ[N-1] = 1; see comment above
}


complex<DOUBLE> Current::gluonAmplitude(){

  return( DOTP(JJJ[0][ngluon-1],JJJ[ngluon-2][0]) 
	  * DOUBLE(2.) * I/pow(sqrt(DOUBLE(2.)),ngluon));
}

DOUBLE Current::checkAmplitude(){
  // check gives bad values only for high multiplicities
  DOUBLE ratio = zero;
  complex<DOUBLE> tree[ngluon];

  for (int i=0; i<ngluon; i++){
    tree[i] =  DOTP(JJJ[0][(ngluon-1+i)%ngluon],JJJ[ngluon-2][i]);
    for (int j=0; j<i; j++){
      DOUBLE tmp = abs(DOUBLE(1.)-tree[i]/tree[j]);
      if (tmp > ratio)
	ratio = tmp;
    }
  }

  return ( ratio );
}

int Current::checkHelsum(const int i, int j,const int hel_i, const int hel_j){

  const int N = (j-i+ngluon) % ngluon + 3;
  const int Helsum_ij = posX(i,j,Helsum) + hel_i + hel_j;


  // returns 0 for "all equal helicity" for N=3
  if ( N == 3 ) return(Helsum_ij % N);

  // exludes a spurious 0 that corresponds for example to ++--
  if( Helsum_ij == 0 )
    return(1);
  // returns 0 for "all equal helicity", "all equal helicity except one"
  
  return ( ( Helsum_ij % N ) * ( Helsum_ij % (N-2) ) );


}

