// $Modified: Mon Oct 11 23:41:54 2010 by uwer $

// vim: set foldmethod=marker:
#include <getopt.h>
#include <iostream>
#include <fstream>
#include <cstdlib>
#include <cmath>
#include <string>
#include <sstream>
#include <sys/resource.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <bitset>

#include "NGluon.h"

#include "mytypes.h"
#include "GZ.h"
#include "GKM.h"
#include "phasespace.h"

#include "analytic.h"
#include "histogram.h"

#ifndef MARK
#define MARK(X) X
#endif

using namespace std;


int ihelp = 0;
int icheck = 0;
int iruntime = 0;
int iruntime_ave = 0;
int iruntime_ana = 0;
int igkmcheck = 0;
int igzcheck = 0;
int imhvcheck = 0;
int iscalecheck = 0;
int iconfcheck = 0;
int nptsarg = 0;
int hstrarg = 0;

int n = 6;
int maxpts = 10;
string hstr="2";



void helpmessage(){
  cout 
    << "\n\n"
    << "ngluon: A test program to illustrate the usage of the NGluon \n"
    << "        library and perform a few basic tests to assure the correct\n"
    << "        working of the library.\n\n"
    << "(c) Simon Badger, Benedikt Biedermann, Peter Uwer, 2010.\n\n"
    << "Usage:\n\n"
    << "Mandatory arguments: Please choose one of the following options \n"
    << "to perform a check for fixed helicity configurations against literature:\n\n"
    << "  --GKMcheck  Run cross checks against [arXiv:0801.2237]\n\n"
    << "  --GZcheck   Run cross checks against [arXiv:0805.2152]\n\n"
    << "  --MHVcheck  Check the MHV configuration against the n-particle\n"
    << "              analytic expression of [hep-ph/0509358]\n\n\n"
    << "To perform accuracy and runtime benchmarks (benchmark options) use:\n\n"
    << "  --scalecheck  Check the error estimate from scaling relations\n"
    << "                (see Fig. 2 of the documentation).\n\n"
    << "  --check       Run accuracy histograms over a number of phase-space points\n"
    << "                (see Fig. 3, 5 and 6 of the documentation).\n\n"
    << "  --confcheck   Check the confidence limit of the finite part\n"
    << "                (see Fig. 4 of the documentation).\n\n"
    << "  --runtime     Runtime estimate for fixed helicity\n\n"
    << "  --runtime_ave Runtime estimate averaged over helicity\n"
    << "                (see Fig. 7 of the documentation).\n\n\n"
    << "Optional arguments to be combined with the benchmark options\n\n"
    << "  --ngluon x to set the number of gluons to x \n\n"
    << "  --npoints x to set the number of phase space points to x \n\n"
    << "            [no effect for --runtime or --runtime_ave]\n\n"
    << "  --hstr x to set the helicty configuration \n"
    << "    use: 0 for (+)^n \n"
    << "         1 for -(+)^(n-1) \n"
    << "         2 for --(+)^(n-2)\n"
    << "         3 for (-+)^(n/2)\n"
    << "         4 for (+-)^(n/2)\n"
    << "    or enter the explicit configuration using +-++-... \n"
    << "    The option --hstr is ignored when combined with --MHVcheck or --runtime_ave\n"
    << "\n\n";
}

void parseOptions(int argc, char **argv){

  static struct option long_options[] =
    {
      {"ngluons", required_argument, 0, 'n'},
      {"npoints", required_argument, 0, 'm'},
      {"hstr", required_argument, 0, 'h'},
      {"check", no_argument, &icheck, 1},
      {"runtime", no_argument, &iruntime, 1},
      {"runtime_ave", no_argument, &iruntime_ave, 1},
      {"runtime_ana", no_argument, &iruntime_ana, 1},
      {"scalecheck", no_argument, &iscalecheck, 1},
      {"confcheck", no_argument, &iconfcheck, 1},
      {"GKMcheck", no_argument, &igkmcheck, 1},
      {"GZcheck", no_argument, &igzcheck, 1},
      {"MHVcheck", no_argument, &imhvcheck, 1},
      {"help", no_argument, &ihelp, 1},
      {0, 0, 0, 0}
    };
  
  while(1){
    /*
     * getopt_long stores the option index here. 
     */
    int option_index = 0;
    int c = getopt_long_only(argc, argv, "n:m:h:",
                         long_options, &option_index);

    /* Detect the end of the options. */
    if ( c == -1 ){
      if (optind < argc){
        cout << "Unrecognized additonal input in commandline: ";
        cout << argv[optind] << endl;
        cout << "Abort programm" << endl;
        exit(1);
      }
      break;
    }
    switch (c){
    case 'n':
      sscanf(optarg, "%d", &n);
      break;
    case 'm':
      sscanf(optarg, "%d", &maxpts);
      nptsarg=1;
      break;
    case 'h':
      char s[40];
      sscanf(optarg, "%s", s);
      hstr = s;
      hstrarg = 1;
      break;
    }
  }
}

class Timer {
  int rs;
  struct rusage rstart, rend;
public:
  Timer(){ start(); }
  void start(){ rs=getrusage(0,&rstart);}
  DOUBLE stop(){
    DOUBLE utime = 0;
    if ( (0 == getrusage(0,&rend) ) && (rs==0) ) {
      utime = rend.ru_utime.tv_sec
	+ rend.ru_utime.tv_usec/1000000.0
	- rstart.ru_utime.tv_sec
	- rstart.ru_utime.tv_usec/1000000.0;
    }
    return(utime);
  }
};

class SysInfo {
  Timer timer;
public:
  ~SysInfo(){
    cout << "# Time used for this run: " << timer.stop() << " seconds."
	 << endl;
  }
};

int cmp(string s, string s1, complex<DOUBLE> a, DOUBLE b,DOUBLE & relmax){
  
  int prec=-4711;
  if ( ( abs(a)!=0 ) && ( abs(b)!=0 ) ) {
    DOUBLE tmp = ( abs(a) - abs(b) ) / 2. / (abs(a)+abs(b));
#ifdef EXTENDED_PREC
    prec = static_cast<int>(log10(abs(to_double(tmp))));
#else
    prec = static_cast<int>(log10(abs(tmp)));
#endif
  }

  if ( prec!=-4711 && prec > (- 5) ){
    cout << s << "      : " << MARK(abs(a)) << endl;
    cout << s << s1 << ": " << MARK(b) << endl;
    return 1;
  } else {
    cout << s << "      : " << abs(a) << endl;
    cout << s << s1 << ": "  << b << endl;
    return 0;
  }
}

void checkMHV(const int ngluon){

  DOUBLE cmEnergy = DOUBLE( ngluon );  
  int helicities[ngluon];
  DOUBLE moms[ngluon][4]; 

  NGluon::sethel(ngluon,2,helicities);

  NGluon loop_amp(ngluon,moms,helicities);
  loop_amp.setVerbosity(NGluon::FULL);
  loop_amp.setScaleTest(true);

  complex<DOUBLE> tree, bub, CC_, RR_, fin_, check;

  FourMomentum<DOUBLE> MOMS[ngluon];

  RandNums rn(ngluon);
  KinCuts kincuts(moms,ngluon,cmEnergy,"jade");
  kincuts.set_deltaS(1e-7);

  if(nptsarg==0) maxpts=1;

  for(int i=0 ; i < maxpts ; i++){

  rn.getRandomNumbers();
  momGen(ngluon, cmEnergy, moms, rn);
  while ( kincuts.check() == 1 ){
    rn.getRandomNumbers();
    momGen(ngluon, cmEnergy, moms, rn);
  }
  
  mom2FourMom(ngluon, moms, MOMS);

  loop_amp.evalAmp();
  tree = loop_amp.getAtree();
  RR_  = loop_amp.getArat();
  fin_ = loop_amp.getAfinite();
      
  check  = RMHV(ngluon,MOMS,loop_amp.getAtree());

  int digi=15;
  cout.precision(digi);
  cout << "A" << ngluon << "_finite\t = " << fin_  << "+/-";
  cout.precision(3);
  cout << loop_amp.getAbsError() << endl;
  cout.precision(digi);
  cout << "RAT" << ngluon << "_num\t = " << RR_ << endl;
  cout << "RAT" << ngluon << "_ana\t = " << check << endl;
  cout.precision(1);
  cout << "Relative accuracy via internal scale check \t= " << loop_amp.getRelErrorRR() << endl;
  cout << "Relative accuracy via external analytic check \t= " << abs((RR_-check)/check) << endl;
  cout.precision(4);

  }
} 


void GZcheck(){

  cout << "\n\n";
  cout << "---------------------------------------------------------------\n";
  cout << "Numerical comparison with values published in: \n";
  cout << "Giele,Zanderighi: \n";
  cout << "On the Numerical Evaluation of One-Loop Amplitudes:\n";
  cout << "The Gluonic Case.\n";
  cout << "JHEP 0806:038,2008. \n";
  cout << "---------------------------------------------------------------\n";
  cout << "\n\n";

  cout.setf(ios_base::scientific, ios_base::floatfield);
  cout.precision(15);
  
  const int nlist[7]={6,7,8,9,10,15,20};
  
  DOUBLE k[20][4];
  int helicities[20];
  int ngluon; 
  
  DOUBLE GZres[5][4];

  for(int ngidx=0; ngidx<7; ngidx++){
    ngluon = nlist[ngidx];
    cout << "-------------------------------"
	 <<"--------------------------------\n";
    cout << "#Number of gluons = " << ngluon << endl;
    cout << "-------------------------------"
	 <<"--------------------------------\n";
    GZsetmom(ngluon,k,GZres);
    NGluon loop_amp(ngluon,k,helicities);
    loop_amp.setMuR(double(ngluon*ngluon));

    DOUBLE res;
    for(int hidx=0;hidx<5;hidx++){
      NGluon::sethel(ngluon,hidx,helicities);
      cout << "Helicities: " << NGluon::helicity2string(ngluon,helicities) 
	   << endl;
      loop_amp.evalAmp();
      cmp("tree     ","(GZ)  ",loop_amp.getAtree(),GZres[hidx][0],res);
      cmp("|Aeps2|  ","(GZ)  ",loop_amp.getAeps2(),GZres[hidx][1],res);
      cmp("|Aeps1|  ","(GZ)  ",loop_amp.getAeps1(),GZres[hidx][2],res);
      cmp("|Afinite|","(GZ)  ",loop_amp.getAfinite(),GZres[hidx][3],res);
    }
  } 
}

void GKMcheck(){
  cout.precision(8);
  cout.setf(ios_base::scientific,ios_base::floatfield);

  cout << "\n\n";
  cout << "---------------------------------------------------------------\n";
  cout << " Cross checking results of GKM [arXiv:0801.2237] " << endl;
  cout << "---------------------------------------------------------------\n";

  complex<DOUBLE> checksum = DOUBLE(0.);

  int check=0;
  int ihel;
  GKM res[10];
  DOUBLE GKM[10][4];
  DOUBLE ctest;
  for(int n=4; n<7;n++){
    cout << "---------------------------------------------------------------\n"
	 << n 
	 << "-point helicity amplitudes " << endl
	 << "---------------------------------------------------------------\n";
    GKMsetmom(n,GKM,res);
    ihel=0;
    while(res[ihel].hel.find("none") == string::npos){
      int h[n];
      NGluon::sethel(n, res[ihel].hel, h);
      NGluon amp(n,GKM,h);
      //      amp.setVerbosity(NGluon::QUIET);
      amp.evalAmp();
      cout << "Helicities: " << res[ihel].hel << endl;
      if (ihel==0 || ihel==1){
	check+=cmp("amp","(GKM) ", amp.getAfinite(),abs(res[ihel].res),ctest);
      } else {
	check+=cmp("amp","(GKM) ", amp.getAfinite()/amp.getAtree(),
	    abs(res[ihel].res),ctest);
	DOUBLE irtest = real( (amp.getAeps2() + amp.getAeps1() )
			      / amp.getAtree() ) ;
	check+=cmp("IR-pole","(ana) ",irtest, - ( amp.IRpoles(-2).real() + amp.IRpoles(-1).real() )
	    ,ctest);
      }
      if ( n==5 ){
	// 5-point amplitudes in [arXiv:0801.2237] are complex conjugated w.r.t. our results
	if ( ihel == 2 ){
	  NGluon::sethel(5,"++---",h);
	  amp.evalAmp();
	  cout << "amp(++---) " << abs( amp.getAfinite()/amp.getAtree()) << endl;
	  check-=1;
	}
	if ( ihel == 3 ){
	  NGluon::sethel(5,"+-+--",h);
	  amp.evalAmp();
	  cout << "amp(+-+--) " << abs( amp.getAfinite()/amp.getAtree()) << endl;
	  check-=1;
	}
      }
      ihel++;
    }
  }

  if(check==0){
  cout << "\n\n";
  cout << "---------------------------------------------------------------\n";
  cout << "  all checks with IR/UV poles and JHEP 0806:038,2008 passed. \n";
  cout << "---------------------------------------------------------------\n";
  cout << "\n\n";
  }
  else{
  cout << "\n\n";
  cout << "---------------------------------------------------------------\n";
  cout << " !!! checks with UV/IR poles and JHEP 0806:038,2008 failed !!! \n";
  cout << " please make sure to compile with scalar integrals \n";
  cout << " using ENABLE_QL=yes in Makefile.inc\n";
  cout << "---------------------------------------------------------------\n";
  cout << "\n\n";


  }
}

DOUBLE rattest(const int nn, DOUBLE moms[][4], const string hstr,
	       NGluon &loop_amp){

  FourMomentum<DOUBLE> MOMS[nn];
  mom2FourMom(nn, moms, MOMS);
  static bool noanalytic=false;
  complex<DOUBLE> check;

  if ( hstr == "0" && nn == 6 ){
    check = R6pppppp(MOMS[0],MOMS[1],MOMS[2],MOMS[3],MOMS[4],MOMS[5]);
    return ( 2.0*abs(loop_amp.getArat()-check)/abs(loop_amp.getArat()+check)+1e-311);
  }

  if ( hstr == "1" && nn == 6 ){
    check = R6mppppp(MOMS[0],MOMS[1],MOMS[2],MOMS[3],MOMS[4],MOMS[5]);
    return ( 2.0*abs(loop_amp.getArat()-check)/abs(loop_amp.getArat()+check)+1e-311 );
  }

  if ( hstr == "3" && nn == 6 ){
    check = R6mpmpmp(MOMS[0],MOMS[1],MOMS[2],MOMS[3],MOMS[4],MOMS[5]);
    return ( 2.0*abs(loop_amp.getArat()-check)/abs(loop_amp.getArat()+check)+1e-311 );
  }

  if (hstr == "2"){
    check = RMHV(nn,MOMS,loop_amp.getAtree());
    return ( 2.0*abs(loop_amp.getArat()-check)/abs(loop_amp.getArat()+check)+1e-311 );
  }

  if(noanalytic==false){
    cout << "# ERROR: no analytic checks are available: returning 0 " << endl;
    noanalytic=true;
  }
  return (0.0);

}


void checkAcc(const int ngluon, const string hstr, const int maxpts){

  cout.setf(ios_base::scientific,ios_base::floatfield);  

#ifdef EXTENDED_PREC
  string prec="ext";
#else
  string prec="dbl";
#endif
  
  double tiny = 1e-311;
  DOUBLE cmEnergy = 1.;
  DOUBLE moms[ngluon][4];
  int helicities[ngluon];
  NGluon::sethel(ngluon,hstr,helicities);
  
  NGluon loop_amp(ngluon,moms,helicities);
  loop_amp.setScaleTest(true);
  RandNums rn(ngluon);

  KinCuts kincuts(moms,ngluon,cmEnergy,"GKM");

#ifdef EXTENDED_PREC
  int accmin = -64;
#else
  int accmin = -16;
#endif
  int accmax = 10;
  int nbins = (accmax-accmin)*3;

  Histogram hist[] = { Histogram("eps2_",accmin,accmax,nbins), 
		       Histogram("eps1_",accmin,accmax,nbins), 
		       Histogram("eps0_",accmin,accmax,nbins), 
		       Histogram("rat_",accmin,accmax,nbins)}; 

  cout << "# -------------------------------------------------------------\n";
  cout 
    << "# "<< ngluon << "-gluon " 
    << NGluon::helicity2string(ngluon, helicities) 
    << " amplitude" << endl
    << "# Accuracy test in " << prec << " precision with " 
    << maxpts << " phase-space points" << endl
    << "# Center of mass energy s = " << cmEnergy << endl
    << "# Kinematic cuts used: " << kincuts.cut << endl
    << "# Seed for random number generation with drand48(): seed = " 
    << rn.seed <<  endl; 
  cout << "# -------------------------------------------------------------\n";

  Timer timer;
  for(int pts=1; pts<=maxpts; pts++){

    rn.getRandomNumbers();
    momGen(ngluon, cmEnergy, moms, rn);

    if ( kincuts.check() ){
      pts-=1;
      kincuts.count++;
      continue;
    }

    if ( loop_amp.CheckPhaseSpacePoint(ngluon,moms) ){
      pts-=1;
      continue;  
    }

    loop_amp.evalAmp();

    hist[0].fill( log10( loop_amp.getIRCheck(-2)) );
    hist[1].fill( log10( loop_amp.getIRCheck(-1)) );
    hist[2].fill( log10( loop_amp.getRelError() + tiny ) );
    hist[3].fill( log10( rattest(ngluon,moms,hstr,loop_amp) ) );

  }

  cout << "# Time used [s]: " << timer.stop() << endl;
  cout << "### DATA " << ngluon << " " 
       << NGluon::helicity2string(ngluon,helicities) << " ###" << endl;

  hist[0].printMultiTable(4,hist);

  cout << "### DATA END ###" << endl;
  
  for (int accmin=0; accmin>=-5; accmin--){
    cout << "# number points with accuracy > " << accmin << " = "; 
    printf("%.4f",hist[2].sumrange( static_cast<double>(accmin),static_cast<double>(accmax))/maxpts*100.);
    cout << "%" << endl;
  }
}

void performance(string mode){

  cout.precision(4);
  cout.setf(ios_base::fixed,ios_base::floatfield);

  int helamps, helicity[n];
  DOUBLE time, moms[n][4];
  Timer timer;

  if (mode == "fix"){
    if (hstrarg==1){
      NGluon::sethel(n,hstr,helicity);
    }
    else{
      NGluon::sethel(n,3,helicity);
    }
  } else if (mode == "ave"){
    cout << "# Average of all helicity combinations" << endl;
  }

  cout << "# Testing " << n << " gluon " 
       << NGluon::helicity2string(n,helicity) 
       << " amplitude (time in ms)" << endl;

  NGluon loop_amp(n,moms,helicity);
  DOUBLE cmEnergy = n;
  RandNums rn(n);
  
  time = DOUBLE(0.);

  KinCuts kincuts(moms,n,cmEnergy,"jade");
  kincuts.set_deltaS(1e-6);
  rn.getRandomNumbers();
  momGen(n, cmEnergy, moms, rn);

  while(kincuts.check()==1){
    rn.getRandomNumbers();
    momGen(n, cmEnergy, moms, rn);
  }
 
  helamps = 1;

  if (mode == "ave")
    for (int i=0; i<n; i++) helamps *= 2;
  
  int pts=0; 
  while (time < 1.){
    for (int hidx = 0; hidx < helamps; hidx++){
      if (mode == "ave"){
	bitset<32> hbin(hidx);
	for(int h=n-1 ; h >= 0 ; h--){
	  helicity[h] = (hbin[h] == 1) ? -1 : 1;
	}
      }
      timer.start();
      loop_amp.evalAmp();
      time += timer.stop();
    }
    pts++;
  }
  time *= 1000./(pts*helamps);

  cout << "### DATA "  << endl; 
  cout << n << " " << time << endl;
  cout << "### DATA END ###" << endl;
}


void checkConfidence(const int ngluon, const string hstr, const int maxpts){
  
  double tiny = 1.e-310;

  DOUBLE cmEnergy = 1.0;
  DOUBLE moms[ngluon][4];
  int helicities[ngluon];
  NGluon::sethel(ngluon,hstr,helicities);

  NGluon loop_amp(ngluon,moms,helicities);
  loop_amp.setScaleTest(true);

  RandNums rn(ngluon);
  KinCuts kincuts(moms,ngluon,cmEnergy,"GKM");

  Histogram confidence("confidence",-16,4,40);

  for(int pts=0; pts<maxpts; pts++){
    
    rn.getRandomNumbers();
    momGen(ngluon, cmEnergy, moms, rn);
    
    if(kincuts.check()){
      pts--;
      kincuts.count++;
      continue;
    }
    
    if ( loop_amp.CheckPhaseSpacePoint(ngluon,moms) ){
      pts-=1;
      continue;  
    }

    loop_amp.evalAmp();
    
    confidence.fill( log10( loop_amp.getRelError() + tiny ) );
  }

  cout << "### DATA " << ngluon << " " 
       << NGluon::helicity2string(ngluon, helicities) << endl;
  
  cout << ngluon << "\t" << confidence.average() 
       << "\t" << confidence.rtmeansqdev() << endl;
  cout << "### DATA END ###" << endl;

}

void checkScaling(const int ngluon, const string hstr, const int maxpts){
  
  DOUBLE cmEnergy = 1.0;
  DOUBLE moms[ngluon][4];
  int helicities[ngluon];
  NGluon::sethel(ngluon,hstr,helicities);

  NGluon loop_amp(ngluon,moms,helicities);
  loop_amp.setScaleTest(true);

  RandNums rn(ngluon);
  KinCuts kincuts(moms,ngluon,cmEnergy,"GKM");
  
  Histogram scaling[] = {
    Histogram("scale test : log(a/b) : 25 bins",-5,5,12*2+1),
    Histogram("scale test : log(a/b) : 49 bins",-5,5,24*2+1),
    Histogram("scale test : log(a)/log(b) : 25 bins",0.5,1.5,12*2+1),
    Histogram("scale test : log(a)/log(b) : 49 bins",0.5,1.5,24*2+1),
    Histogram("scale test : log(a)/log(b) : 97 bins",0.5,1.5,48*2+1)
  };
  

  cout << "#!!! errors < 1e-15 ignored !!!#" << endl; 
  for(int pts=0; pts<maxpts; pts++){
    
    rn.getRandomNumbers();
    momGen(ngluon, cmEnergy, moms, rn);
    
    if ( kincuts.check() ){
      pts--;
      kincuts.count++;
      continue;
    }
   
    if ( loop_amp.CheckPhaseSpacePoint(ngluon,moms) ){
      pts-=1;
      continue;  
    }
 
    loop_amp.evalAmp();
    
    DOUBLE a = rattest(ngluon,moms,hstr,loop_amp);
    DOUBLE b = loop_amp.getRelErrorRR();

    if(log10(a) < -15.0){
      continue;
    }

    scaling[0].fill( (log10(a)-log10(b)) );
    scaling[1].fill( (log10(a)-log10(b)) );
    scaling[2].fill( (log10(a)/log10(b)) );
    scaling[3].fill( (log10(a)/log10(b)) );
    scaling[4].fill( (log10(a)/log10(b)) );
  }
  
  DOUBLE binaverage;

  for (int k=0 ; k < 5 ; k++){
    cout << "### DATA " << k << " " << ngluon << " " 
      << NGluon::helicity2string(ngluon, helicities) << endl;
    cout << "#bins\t" << "\t" << scaling[k].getname() << endl;
 
    binaverage = 0.0;
    for (int j=0 ; j < scaling[k].getbins() ; j++){
      cout << scaling[k].getCenter(j) << "\t" << scaling[k].getpts(j) << endl;
      binaverage += scaling[k].getCenter(j)*scaling[k].getpts(j)/scaling[k].getentries();
    }
    cout << "### DATA END ###" << endl;
    cout << "### Average from bins = " << binaverage << " ###"<< endl;
    cout << "### Average from data = " << scaling[k].average() << " ###"<< endl;
  }

}

static FloatingPointArithmetic fpa;

int main(int argc, char **argv){

  SysInfo sysinfo;
  parseOptions(argc,argv);

  if (icheck) {
    checkAcc(n, hstr, maxpts);
    return(0);
  }

  if (imhvcheck) {
    cout.setf(ios_base::scientific,ios_base::floatfield);
    checkMHV(n);
    return(0);
  }

  if (iruntime) {
    performance("fix");
    return(0);
  }

  if (iruntime_ave) {
    performance("ave");
    return(0);
  }

  if (igkmcheck) {
    GKMcheck();
    return(0);
  }

  if (igzcheck) {
    GZcheck();
    return(0);
  }

  if (iscalecheck) {
    cout.setf(ios_base::scientific,ios_base::floatfield);
    checkScaling(n, hstr, maxpts);
    return(0);
  }

  if (iconfcheck) {
    cout.setf(ios_base::scientific,ios_base::floatfield);
    checkConfidence(n, hstr, maxpts);
    return(0);
  }

  helpmessage();

}
