/* $Modified: Thu Nov 11 16:36:55 2010 by uwer $ */
// vim: set ft=cpp:
#ifndef HISTOGRAM_H_
#define HISTOGRAM_H_
#include <iostream>
#include "mytypes.h"

class Histogram {
 private:
  double xmin,xmax,binsize;
  int bincount, underflow, overflow, entries;
  std::string histname;
  double avg, sd;
  int *dataptr;
 public:

  Histogram(std::string name_, double xmin_, double xmax_,  
	    int bincount_) : xmin(xmin_), xmax(xmax_),bincount(bincount_),
    histname(name_) {
    binsize = (xmax-xmin) / bincount;
    dataptr = new int[bincount];
    reset();
  };
    
  ~Histogram(){
    std::cout << "# ------------------------------------------ " << std::endl;
    std::cout << "# Statistics for " << histname << std::endl;
    std::cout << "# Entries: " << entries << std::endl;
    std::cout << "# Underflow: " << underflow << std::endl;
    std::cout << "# Overflow: " << overflow << std::endl;
    std::cout << "# Average: " <<  average() << std::endl;
    std::cout << "# Root Mean Square Deviation: " <<  rtmeansqdev() << std::endl;
    std::cout << "# ------------------------------------------ " << std::endl;
    delete [] dataptr;
  };
  
  std::string getname(){return(histname);};

  void reset(){
    underflow = overflow = 0; 
    entries = 0; 
    avg = sd = 0;
    for (int i=0; i<bincount; i++)
      dataptr[i] = 0;
  }
  
  double getCenter(int bin) {
    return( xmin + binsize/2.0 + bin*binsize ); 
  }
  
  int getpts(const int i){
    return (dataptr[i]);
  }

  int getentries(){
    return(entries);
  }


  int getbins(){
    return(bincount);
  }

  int getBin(double x) {
    if ( ( x < xmin ) || (x > xmax) ) {
      return (-1);
    }
    return(static_cast<int>( trunc( ( x - xmin ) / binsize ) ) );
  }

  void fill(const DOUBLE x_){
    double x;
#ifdef EXTENDED_PREC
    x = to_double(x_);
#else
    x = x_;
#endif
    int bin = getBin(x);
 
    if ( bin < 0 ) {
      ( x < xmin ) ? underflow++ : overflow++;
      return;
    } 

    entries++;
    dataptr[bin]++;
    avg += x;
    sd += x*x;
  }
  
  double average(){
    return (avg/entries);
  }
  
  double rtmeansqdev(){

    double nn=static_cast<double>(entries); 
    
    return (sqrt( 
	( sd - avg*avg/nn ) /(nn-1)/nn 
	));
  }

double sumrange(const double minlim, const double maxlim){

  int minbin = getBin(minlim);
  int maxbin = getBin(maxlim);
  int sum = 0;

  for(int k= minbin ; k < maxbin ; k++){
    sum += dataptr[k];
  }

  return ( double(sum) );
}

void printMultiTable(int n, Histogram h[] ){

  int nbin = h[0].getbins();
  for(int i=1; i< n; i++){
    if ( nbin != h[i].getbins() ){
      std::cout << "Histgrams differ in length..." << std::endl;
      return;
    }
  }

  std::cout << "#bins\t";
  for(int i=0; i< n; i++)
    std::cout << "\t" << h[i].getname();
  std::cout << std::endl;
  
  for (int j=0; j < nbin; j++){
    std::cout << h[0].getCenter(j);
      for(int i=0; i< n; i++)
	std::cout << "\t" << h[i].getpts(j);
    std::cout << std::endl;
  }
}

};

#endif
