// $Modified: Tue Nov  2 23:21:39 2010 by uwer $
#include <cmath>
#include <cstdlib>
#include <iostream>
#include "phasespace.h"

using namespace std;

#define MAXN n

static inline DOUBLE lambda(const DOUBLE &a, const DOUBLE &b, const DOUBLE &c){
  return a*a + b*b + c*c - DOUBLE(2.)*a*b - DOUBLE(2.)*b*c - DOUBLE(2.)*a*c;
} 

static inline void boost(const DOUBLE m, const DOUBLE p[4], DOUBLE mom[4]){

  //  m = (m == 0.) ? p[0]*p[0]-p[1]*p[1]-p[2]*p[2]*p[3]*p[3] : m ;
  const DOUBLE gamma = p[0] / m; 
  const DOUBLE invp0 = DOUBLE(1.) / p[0];
 
  const DOUBLE beta1 = p[1] * invp0;
  const DOUBLE beta2 = p[2] * invp0;
  const DOUBLE beta3 = p[3] * invp0;

  const DOUBLE help = beta1 * mom[1] + beta2 * mom[2] + beta3 * mom[3];
  const DOUBLE fac = ( gamma / ( DOUBLE(1.) + gamma ) * help + mom[0] ) * gamma;

  mom[0] = ( mom[0] + help ) * gamma ;
  mom[1] += fac * beta1;
  mom[2] += fac * beta2;
  mom[3] += fac * beta3;

}

int eventn(const unsigned int n, const DOUBLE sqrts, const double x[], 
           const DOUBLE masses[], DOUBLE kout[][4], DOUBLE& jacobi){
  
  const DOUBLE zero(0.),one(1.),two(2.);
  const DOUBLE PI(SETVAL(3.14159265358979323846264338327950288419716939937510582097494459230));

  /* 
   * Routine to generate a n-parton event with center-of-mass
   * energy squared given by s and the parton masses set through
   * the 4-momenta in kout[].
   * The momentum configuration is returned in kout[], the jacobian for
   * the transformation is returned through jacobi.
   * x should contain 3*n-4 random numbers
   * Algorithm is described in Byckling & Kajantie
   */
  
  static const DOUBLE TwoPi = two * PI;
  int err = 0;

  struct {
    DOUBLE m;
    DOUBLE MM;
    DOUBLE MMQ;
    DOUBLE mu;
    DOUBLE P;
  } pinfo[MAXN];

  //int irn = 0;
  //#define XRANDOM x[irn++]
#define XRANDOM  *(x++)


  DOUBLE msum=0;
  for ( unsigned int i = 0; i < n; i++) {
    pinfo[i].m = masses[i];
    msum += pinfo[i].m;
    pinfo[i].mu = msum;
  };
  
  /*
   * Set the invariant masses MMQ(i) = (p_0+...+p_i)^2
   */     
  pinfo[n-1].MM  = sqrts; 
  pinfo[n-1].MMQ = sqrts*sqrts;
  pinfo[0].MM    = pinfo[0].m;
  pinfo[0].MMQ   = pinfo[0].MM * pinfo[0].MM;

  for (unsigned int i = n-2; i > 0; i--) {
    pinfo[i].MM = pinfo[i].mu + XRANDOM * ( pinfo[i+1].MM - pinfo[i+1].mu );
    pinfo[i].MMQ = pinfo[i].MM * pinfo[i].MM;
  }

  /*
   * Calculate the three momenta and the jacobian
   */
  pinfo[1].P = pinfo[1].m == zero ? ( pinfo[1].MM - pinfo[0].MM ) * 
    ( pinfo[1].MM + pinfo[0].MM ) / ( two * pinfo[1].MM ) :
      sqrt( lambda(pinfo[1].MMQ, pinfo[0].MMQ, pinfo[1].m*pinfo[1].m ) ) 
    / ( two * pinfo[1].MM );

  jacobi = pow(TwoPi, static_cast<int>(n)-1) / ( two * sqrts ) * pinfo[1].P;


  for( unsigned int i = 2; i < n; i++ ) {
    pinfo[i].P = 
      ( pinfo[i].m == zero ? 
	( pinfo[i].MM - pinfo[i-1].MM ) * ( pinfo[i].MM + pinfo[i-1].MM ) 
	/ ( two * pinfo[i].MM ) :
	sqrt( lambda(pinfo[i].MMQ, pinfo[i-1].MMQ, pinfo[i].m*pinfo[i].m ) ) 
	/ ( two * pinfo[i].MM ) );
    jacobi *= pinfo[i].P * ( pinfo[i].MM - pinfo[i].mu );
 }

  /*
   * Finally reconstruct the event in the overall CMS:
   *
   * Calculate first the 4-momenta pset(rho,i) in the restframe of
   * p_1+...+p_i:
   */
  for(unsigned int i = 1; i < n; i++) {
#ifdef USE_SINCOS
    DOUBLE sinphi_i, cosphi_i;
    sincos(TwoPi * XRANDOM, &sinphi_i, &cosphi_i);
#else
    /*
     *  Note: For phi we have to calculate the cosine explicitly!!!!
     *  dsqrt(1-sin^2) would only give cos() > 0
     */
    DOUBLE phi_i = TwoPi * XRANDOM;
    DOUBLE sinphi_i = sin( phi_i );
    DOUBLE cosphi_i = cos( phi_i );
#endif
    DOUBLE tmp = XRANDOM;
    DOUBLE costheta_i = two * tmp - one;
    DOUBLE sintheta_i = two * sqrt( tmp * ( one - tmp ) );
    //DOUBLE sintheta_i = sqrt(one - costheta_i*costheta_i);

    kout[i][1] = pinfo[i].P * cosphi_i*sintheta_i;
    kout[i][2] = pinfo[i].P * sinphi_i*sintheta_i;
    kout[i][3] = pinfo[i].P * costheta_i;
    kout[i][0] = pinfo[i].m == 0. ? pinfo[i].P : 
      sqrt( pinfo[i].m*pinfo[i].m + pinfo[i].P*pinfo[i].P );
    
  }
  
  kout[0][0] = pinfo[1].MM - kout[1][0];
  kout[0][1] = - kout[1][1];
  kout[0][2] = - kout[1][2];
  kout[0][3] = - kout[1][3];

  DOUBLE bhelp[4];
  for (unsigned int i = 2; i < n; i++) {
    bhelp[0] = pinfo[i].MM-kout[i][0]; 
    bhelp[1] = -kout[i][1]; 
    bhelp[2] = -kout[i][2]; 
    bhelp[3] = -kout[i][3];
    for (unsigned int j = 0; j < i; j++){
      boost(pinfo[i-1].MM,bhelp,&kout[j][0] );
    }
  }  
  return err;
}

void momGen(const int ngluon, const DOUBLE sqrts, DOUBLE kk[][4],
	    RandNums& rn){
  
  const DOUBLE zero(0.),two(2.);
  const int NN = ngluon-2; 
  DOUBLE ms[NN];
  for (int i=0; i<NN; i++){
    ms[i] = zero;
  }
  DOUBLE wgt;
  
  eventn(NN, sqrts, rn.xx, ms, &kk[2], wgt);
  kk[0][0] = -sqrts/two; kk[0][1] = zero; kk[0][2]=zero; kk[0][3] = -sqrts/two;
  kk[1][0] = -sqrts/two; kk[1][1] = zero; kk[1][2]=zero; kk[1][3] =  sqrts/two;
}

void momGen(FourMomentum<DOUBLE> *KK, const int ngluon,
	    const DOUBLE cmEnergy, RandNums& rn){
  
  DOUBLE kk[ngluon][4];
  momGen(ngluon, cmEnergy, kk, rn);
  mom2FourMom(ngluon, kk, KK);
}

void mom2FourMom(const int ngluon, DOUBLE kk[][4], FourMomentum<DOUBLE> KK[]){
  for (int i = 0; i < ngluon; i++){
    KK[i].setFourMomentum(kk[i]);
  }
}

void KinCuts::info(string s){
  cout << "# KINCUTS: " << s  <<endl;
}

void KinCuts::info(string s, DOUBLE tmp){
  cout << "# KINCUTS: " << s << " " << tmp <<endl;
}

bool KinCuts::check() {
  return (this->*function)();
}

void KinCuts::set_cmEnergy(const DOUBLE cmEnergy_){
  cmEnergy = cmEnergy_;
}

void KinCuts::set_pTMin(const DOUBLE pTMin_){ 
  pTMin = pTMin_;
  info("  -> set pTMin = cmEnergy *",pTMin);
}

void KinCuts::set_etaMax(const DOUBLE etaMax_){ 
  etaMax = etaMax_;
  info("  -> set etaMax = ",etaMax);
}

void KinCuts::set_RijMin(const DOUBLE RijMin_){ 
  RijMin = RijMin_;
  info("  -> set R_ijMin = ",RijMin);
}

void KinCuts::set_deltaS(const DOUBLE deltaS_){
  deltaS = deltaS_;
  info("  -> set deltaS = ",deltaS);
}

DOUBLE KinCuts::rapidity(DOUBLE p[4]){
  // with respect to z-axis
  return ( 0.5 * log((p[0] + p[3])/(p[0] - p[3])) );
}

DOUBLE KinCuts::transvMom(DOUBLE p[4]){
  return (sqrt(p[1]*p[1] + p[2]*p[2]));
}

DOUBLE KinCuts::phi(DOUBLE p[4]){
  DOUBLE phi = acos( p[1]/(sqrt(p[1]*p[1] + p[2]*p[2])) );
  if (p[2] > 0.0){
    return ( phi );
  }
  else {
    return (phi + M_PI);
  }
}

bool KinCuts::etaPhiPT(){
  for (int i=2; i<nn; i++){
    pT[i]  = transvMom(moms[i]);
    if (pT[i] < pTMin)
      return (1);
    eta[i] = rapidity(moms[i]);
    if (abs(eta[i]) > etaMax)
      return (1);
    Phi[i] = phi(moms[i]);
    for (int j=2; j<i; j++){
      deltaR = sqrt( pow((eta[i]-eta[j]),2) + pow((Phi[i]-Phi[j]),2));
      if (deltaR < RijMin) 
	return (1);
    }
  }
  return (0);
}

inline DOUBLE KinCuts::dot(DOUBLE p[4], DOUBLE q[4]){
  return (p[0]*q[0] - (p[1]*q[1] + p[2]*q[2] + p[3]*q[3] ) ); 
}

bool KinCuts::jade(){
  for (int i=0; i<nn; i++){
    for (int j=i+1;j<nn;j++){
      if ( fabs( 2.0 * dot(moms[i],moms[j]) ) < deltaS)
	return (1);
    }
  }
  return (0);
}
